/*
   Alfi-plotter image drawer, text typer.

   Written by ond <taxman@bimbo.fjfi.cvut.cz> at 09121999th1848.
 */

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include "alfi.h"

struct alfi_image_st
  {
    char *data;
    int width, height;
  };

void
die (const char *msg)
{
  printf ("%s\n", msg);
  exit (-1);
}

int
find_nearest_dot (struct alfi_image_st *image, int color,
		  int *dx, int *dy, int x, int y)
{
  int i, j, k, n, n_x, n_y;

  n = -1;
  for (j = 0; j < image->height; j++)
    for (i = 0; i < image->width; i++)
      {
	if (image->data[i + j * image->width] == color)
	  {
	    k = (i - x) * (i - x) + (j - y) * (j - y);
	    if (n == -1 || k < n)
	      {
		n = k;
		n_x = i;
		n_y = j;
	      }
	  }
      }

  if (n == -1)
    return -1;

  *dx = (n_x - x);
  *dy = (n_y - y);

  return 0;
}

int
get_longest_line (struct alfi_image_st *image, int color,
		  int *lx, int *ly, int x, int y)
{
  int i, l, l_dir, dir, x_pos, y_pos;
  char dirs[8][2] =
  {
    {0, -1},
    {1, -1},
    {1, 0},
    {1, 1},
    {0, 1},
    {-1, 1},
    {-1, 0},
    {-1, -1},
  };

  l = 0;
  l_dir = 0;

  for (dir = 0; dir < 8; dir++)
    {
      i = 0;

      for (;;)
	{
	  x_pos = x + dirs[dir][0] * i;
	  y_pos = y + dirs[dir][1] * i;

	  if (x_pos < 0 || x_pos >= image->width || y_pos < 0 || y_pos >= image->height)
	    break;

	  if (image->data[x_pos + y_pos * image->width] != color)
	    break;

	  i++;
	}

      if (i > l)
	{
	  l = i;
	  l_dir = dir;
	}
    }

  if (!l)
    return -1;

  for (i = 0; i < l; i++)
    {
      image->data[x + dirs[l_dir][0] * i +
		  (y + dirs[l_dir][1] * i) * image->width] = 0x00;
    }

  i--;

  *lx = i * dirs[l_dir][0];
  *ly = i * dirs[l_dir][1];

  return 0;
}

void
print_image (char *image_data, int pen_x, int pen_y, int size_x, int size_y,
	     int magnify_x, int magnify_y, int color)
{
  int dx, dy, lx, ly;
  struct alfi_image_st image;

  image.data = image_data;
  image.width = size_x;
  image.height = size_y;

  pen_x /= magnify_x;
  pen_y /= magnify_y;

  while (!find_nearest_dot (&image, color, &dx, &dy, pen_x, pen_y))
    {
      if (abs (dx) > 1 || abs (dy) > 1)
	alfi_pen (FALSE);

      alfi_move (dx * magnify_x, dy * magnify_y);

      pen_x += dx;
      pen_y += dy;

      if (get_longest_line (&image, color, &lx, &ly, pen_x, pen_y))
	break;

      pen_x += lx;
      pen_y += ly;

      alfi_pen (TRUE);
      alfi_line (magnify_x * lx, magnify_y * ly);
    }

  alfi_pen (FALSE);
}

void
bitmap_2_image (char *dst, char *letter, int lines)
{
  int i, j;

  for (j = 0; j < lines; j++)
    for (i = 0; i < 8; i++)
      dst[i + j * 8] = letter[j] & (0x80 >> i) ? 1 : 0;
}

int
textprint (char *str, int len, char *font, int magnify_x, int magnify_y)
{
  char buff[256];
  int x_pos, y_pos;

  x_pos = 0;
  y_pos = 0;

  while (len--)
    {
      switch (*str)
	{
	case 0x0A:
	  y_pos++;
	  alfi_locate (0, y_pos * 16 * magnify_y);
	  break;
	case 0x0D:		/* forbiden */
	  break;
	default:
	  x_pos++;
	  bitmap_2_image (buff, font + 16 * (*str++), 16);
	  print_image (buff, 0, alfi_y % (16 * magnify_y), 8, 16, magnify_x, magnify_y, 1);
	  alfi_locate (x_pos * 8 * magnify_x, alfi_y);
	  break;
	}
    }
  alfi_locate (0, alfi_y + 16);
}

int
main (int argc, char **args)
{
  int fd, i;
  char *font, *data;

  printf ("Alfi typewriter 1.0\n");
  if (argc != 3)
    {
      printf ("usage: atype textfile fontfile\n");
      return 0;
    }

  printf ("printing file %s using font table %s\n", args[1], args[2]);

  fd = open (args[2], O_RDONLY);
  if (fd == -1)
    die ("cannot open font file");

  font = (char *) malloc (4096);
  read (fd, font, 4096);
  close (fd);

  fd = open (args[1], O_RDONLY);
  if (fd == -1)
    die ("cannot open text file");

  i = lseek (fd, 0, SEEK_END);
  lseek (fd, 0, SEEK_SET);
  data = (char *) malloc (i);
  read (fd, data, i);
  close (fd);

  textprint (data, i, font, 1, 1);

  free (data);
  free (font);

  return 0;
}
